﻿using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration.GrammarEnumeration;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  public class Grammar {

    public Symbol StartSymbol;

    #region Symbols
    public VariableSymbol Var;

    public NonterminalSymbol Expr;
    public NonterminalSymbol Term;
    public NonterminalSymbol Factor;

    public TerminalSymbol Addition;
    public TerminalSymbol Multiplication;
    #endregion


    public Grammar(string[] variables) {
      #region Define Symbols
      Var = new VariableSymbol("var", variables);

      Expr = new NonterminalSymbol("Expr");
      Term = new NonterminalSymbol("Term");
      Factor = new NonterminalSymbol("Factor");

      Addition = new TerminalSymbol("+");
      Multiplication = new TerminalSymbol("*");
      #endregion


      #region Production ruless
      StartSymbol = Expr;

      Expr.AddProduction(Term, Expr, Addition);
      Expr.AddProduction(Term);

      Term.AddProduction(Factor, Term, Multiplication);
      Term.AddProduction(Factor);

      Factor.AddProduction(Var);
      #endregion
    }

    public int CalcHashCode(SymbolString sentence) {
      Debug.Assert(sentence.Any(), "Trying to evaluate empty sentence!");
      Debug.Assert(sentence.All(s => s is TerminalSymbol), "Trying to evaluate symbol sequence with nonterminalsymbols!");

      Stack<TerminalSymbol> parseStack = new Stack<TerminalSymbol>(sentence.OfType<TerminalSymbol>());

      int[] subtreeHashes = GetSubtreeHashes(parseStack.Pop(), parseStack);
      return AggregateHashes(subtreeHashes);
    }

    private int[] GetSubtreeHashes(TerminalSymbol currentSymbol, Stack<TerminalSymbol> parseStack) {
      List<int> childHashes = null;

      if (Var.VariableTerminalSymbols.Contains(currentSymbol)) {
        childHashes = currentSymbol.StringRepresentation.GetHashCode().ToEnumerable().ToList();

      } else if (ReferenceEquals(currentSymbol, Multiplication)) { // MULTIPLICATION
        childHashes = new List<int>();

        // First subtree
        TerminalSymbol firstSubtreeRoot = parseStack.Pop();
        int[] firstSubtreeHashes = GetSubtreeHashes(firstSubtreeRoot, parseStack);

        if (ReferenceEquals(firstSubtreeRoot, Multiplication)) {
          childHashes.AddRange(firstSubtreeHashes);
        } else {
          childHashes.Add(AggregateHashes(firstSubtreeHashes));
        }

        // Second subtree
        TerminalSymbol secondSubtreeRoot = parseStack.Pop();
        int[] secondSubtreeHashes = GetSubtreeHashes(secondSubtreeRoot, parseStack);

        if (ReferenceEquals(secondSubtreeRoot, Multiplication)) {
          childHashes.AddRange(secondSubtreeHashes);
        } else {
          childHashes.Add(AggregateHashes(secondSubtreeHashes));
        }

        // Sort due to commutativity
        childHashes.Sort();



      } else if (ReferenceEquals(currentSymbol, Addition)) { // ADDITION
        HashSet<int> uniqueChildHashes = new HashSet<int>();

        TerminalSymbol firstSubtreeRoot = parseStack.Pop();
        int[] firstSubtreeHashes = GetSubtreeHashes(firstSubtreeRoot, parseStack);

        // First subtree
        if (ReferenceEquals(firstSubtreeRoot, Addition)) {
          uniqueChildHashes.UnionWith(firstSubtreeHashes);
        } else {
          uniqueChildHashes.Add(AggregateHashes(firstSubtreeHashes));
        }

        // Second subtree
        TerminalSymbol secondSubtreeRoot = parseStack.Pop();
        int[] secondSubtreeHashes = GetSubtreeHashes(secondSubtreeRoot, parseStack);

        if (ReferenceEquals(secondSubtreeRoot, Addition)) {
          uniqueChildHashes.UnionWith(secondSubtreeHashes);
        } else {
          uniqueChildHashes.Add(AggregateHashes(secondSubtreeHashes));
        }

        childHashes = uniqueChildHashes.ToList();
        childHashes.Sort();
      }

      Debug.Assert(childHashes != null, "Sentence not hashable!");
      return childHashes.ToArray();
    }


    private int AggregateHashes(IEnumerable<int> hashes) {
      return hashes.Aggregate(0, (result, ti) => ((result << 5) + result) ^ ti.GetHashCode());
    }
  }
}
